/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package SoarBridge;

import Simulation.SimulationRobot;
import Simulation.SimulationCreature;
import Simulation.SimulationCreature.StakeholderType;
import Simulation.SimulationItem;
import Simulation.FuelSensorReadings;
import Simulation.SensorType;
import java.io.File;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import java.util.Set;
import org.apache.log4j.Logger;
import sml.Agent;
import sml.Identifier;
import sml.Kernel;
import ws3dproxy.model.Leaflet;
import ws3dproxy.model.Thing;
import ws3dproxy.util.Constants;

/**
 *
 * @author Danilo
 */
public class SoarBridge
{
    // Log Variable
    Logger logger = Logger.getLogger(SoarBridge.class.getName());

    // SML variables
    Kernel kernel = null;
    Agent agent = null;
    Identifier inputLink = null;

    // Entity Variables
    Identifier creature;
    Identifier creatureSensor;
    Identifier creatureParameters;
    Identifier creaturePosition;
    Identifier creatureMemory;
    Identifier creatureGoal;

    // Ordinary Variables
    String agentName;
    File productionPath;

    /**
     * Constructor class
     * @param _agentName Agent Name
     * @param _productionPath Production Path
     * @throws SoarBridgeException
     */
    public SoarBridge(String _agentName, File _productionPath, Boolean startSOARDebugger, String soarDebuggerPath, Integer soarDebuggerPort) throws SoarBridgeException
    {
        try
        {
            // Checking parameters
            if (_agentName == null || _productionPath == null)
            {
                throw new SoarBridgeException("Parameters are null");
            }

            // Inicial variables
            agentName = _agentName;
            productionPath = _productionPath;

            // create Soar kernel And Agent
            kernel = Kernel.CreateKernelInNewThread();
            checkForKernelOrAgentError();
            agent = kernel.CreateAgent(agentName);
            checkForKernelOrAgentError();

            // Load some productions
            agent.LoadProductions(productionPath.getAbsolutePath());
            checkForKernelOrAgentError();

            inputLink = agent.GetInputLink();

            // Initialize enteties
            creature = null;

            // Debugger line
            if (startSOARDebugger)
            {
                agent.SpawnDebugger(soarDebuggerPort);
            }
        }
        catch (Exception e)
        {
            logger.error("Error while creating SOAR Kernel",e);
            throw new SoarBridgeException("Error while Creating SOAR Bridge");
        }
    }

    /**
     * Finalize the SOAR Kernel
     * @throws SoarBridgeException
     */
    public void finalizeKernel() throws SoarBridgeException
    {
        try
        {
            if (kernel != null)
            {
                kernel.Shutdown();
            }
        }
        catch (Exception e)
        {
            logger.error("Error while creating SOAR Kernel",e);
            throw new SoarBridgeException("Error while Creaing SOAR Bridge");
        }
    }

    /***
     * Clear All Entities in Input Link
     * @throws SoarBridgeException
     */
    public void clearEntitiesToNextSimulation() throws SoarBridgeException
    {
        try
        {
            if (agent != null)
            {
                if (creatureSensor != null && creaturePosition != null && creatureParameters != null)
                {
                    agent.DestroyWME(creatureSensor);
                    agent.DestroyWME(creaturePosition);
                    agent.DestroyWME(creatureParameters);
                }

                checkForKernelOrAgentError();
            }
        }
        catch (Exception e)
        {
            logger.error("Error while creating SOAR Kernel",e);
            throw new SoarBridgeException("Error while Creaing SOAR Bridge");
        }
    }


    /**
     * Create New Entity in SOAR Kernel
     * @param categoryType
     * @param parameters
     * @throws SoarBridgeException
     */
    public void setupStackHolder(StakeholderType entityType, SimulationCreature parameter) throws SoarBridgeException
    {
        try
        {
            if (agent != null)
            {
                switch (entityType)
                {
                    case CREATURE:
                        SimulationRobot creatureParameter = (SimulationRobot)parameter;

                        // Create creature
                        if (creature == null)
                        {                               
                            // Initialize Creature Ententy
                            creature = agent.CreateIdWME(inputLink, "CREATURE");

                            // Initialize Creature Memory
                            creatureMemory = agent.CreateIdWME(creature, "MEMORY");
                        }

                        // Set Creature Parameters
                        Calendar lCDateTime = Calendar.getInstance();                        
                        
                        String query = ProcessLeaflet(creatureParameter.leafletList); 
                        int completed = 0;
                        
                        if(query.isEmpty())
                            completed = 1;
                        else
                            completed = 0;
                        
                        //System.out.println("Resulesta de completo = "+completed);
                        
                        creatureParameters = agent.CreateIdWME(creature, "PARAMETERS");
                        agent.CreateFloatWME(creatureParameters, "MINFUEL", 400);
                        agent.CreateFloatWME(creatureParameters, "TIMESTAMP", lCDateTime.getTimeInMillis());
                        agent.CreateStringWME(creatureParameters,"GOAL",query);
                        agent.CreateIntWME(creatureParameters,"COMPLETEDLEAFLET", completed);
                        
                        // Setting creature Position
                        creaturePosition = agent.CreateIdWME(creature, "POSITION");
                        agent.CreateFloatWME(creaturePosition, "X", creatureParameter.getPosition().getX());
                        agent.CreateFloatWME(creaturePosition, "Y", creatureParameter.getPosition().getY());
                        
                        // Set creature sensors
                        if (creatureParameter.getCountSensorsAvailable() > 0)
                        {
                            // Create Sensor Node
                            creatureSensor = agent.CreateIdWME(creature, "SENSOR");

                            // Get Sensor Enabled
                            List<SensorType> enabledSensorTypeList = creatureParameter.getSensorsEnabled();

                            // Check if we have active sensors in creature
                            if (enabledSensorTypeList != null && enabledSensorTypeList.size() > 0)
                            {
                                for (int sensorTypeCount=0; sensorTypeCount < enabledSensorTypeList.size(); sensorTypeCount++)
                                {
                                    // Check wich sensors are actived
                                    switch (enabledSensorTypeList.get(sensorTypeCount))
                                    {
                                        case FUEL:
                                        {
                                            Identifier fuel = agent.CreateIdWME(creatureSensor, "FUEL");

                                            FuelSensorReadings sensorReadings = (FuelSensorReadings) creatureParameter.getSensorHandler(SensorType.FUEL).getSensorReadings();

                                            if (sensorReadings !=  null)
                                            {
                                                agent.CreateFloatWME(fuel, "VALUE",sensorReadings.getCurrentFuel());
                                                
                                                //System.out.println("Gasolina = "+ sensorReadings.getCurrentFuel());
                                            }
                                        }
                                        break;

                                        case VISUAL:
                                        {
                                            Identifier visual = agent.CreateIdWME(creatureSensor, "VISUAL");
                                            List<Thing> thingsList = (List<Thing>)(creatureParameter.getSensorHandler(SensorType.VISUAL).getSensorReadings());

                                            if (thingsList != null)
                                            {
                                                for (int thingsInSensorCount = 0 ; thingsInSensorCount < thingsList.size(); thingsInSensorCount++)
                                                {
                                                    Identifier entity = agent.CreateIdWME(visual, "ENTITY");
                                                    Thing t = thingsList.get(thingsInSensorCount);
                                                    agent.CreateFloatWME(entity, "DISTANCE", GetGeometricDistanceToCreature(t.getX1(),t.getY1(),t.getX2(),t.getY2(),creatureParameter.getPosition().getX(),creatureParameter.getPosition().getY()));                                                    
                                                    agent.CreateFloatWME(entity, "X", t.getX1());
                                                    agent.CreateFloatWME(entity, "Y", t.getY1());
                                                    agent.CreateFloatWME(entity, "X2", t.getX2());
                                                    agent.CreateFloatWME(entity, "Y2", t.getY2());                            
                                                    agent.CreateStringWME(entity, "TYPE", SimulationItem.getItemTypeFromThingCategory(t.category).toString());
                                                    agent.CreateStringWME(entity, "NAME", t.getName());
                                                    agent.CreateStringWME(entity, "COLOR",Constants.getColorName(t.getMaterial().getColor()));                                                    
                                                }
                                            }
                                        }
                                        break;
                                    }
                                }
                            }
                        }

                        break;
                }

                agent.Commit();

                checkForKernelOrAgentError();
            }
        }
        catch (Exception e)
        {
            logger.error("Error while creating SOAR Kernel",e);
            throw new SoarBridgeException("Error while Creating SOAR Bridge");
        }
    }

    private double GetGeometricDistanceToCreature(double x1, double y1, double x2, double y2, double xCreature, double yCreature)
    {
          float squared_dist = 0.0f;
          double maxX = Math.max(x1, x2);
          double minX = Math.min(x1, x2);
          double maxY = Math.max(y1, y2);
          double minY = Math.min(y1, y2);

          if(xCreature > maxX)
          {
            squared_dist += (xCreature - maxX)*(xCreature - maxX);
          }
          else if(xCreature < minX)
          {
            squared_dist += (minX - xCreature)*(minX - xCreature);
          }

          if(yCreature > maxY)
          {
            squared_dist += (yCreature - maxY)*(yCreature - maxY);
          }
          else if(yCreature < minY)
          {
            squared_dist += (minY - yCreature)*(minY - yCreature);
          }

          return Math.sqrt(squared_dist);
    }

    /**
     * Run SOAR Simulation
     * @throws SoarBridgeException
     */
    public void runSimulation() throws SoarBridgeException
    {
        try
        {
            if (agent != null)
            {
                agent.RunSelfTilOutput();
                checkForKernelOrAgentError();
            }
        }
        catch (Exception e)
        {
            logger.error("Error while creating SOAR Kernel",e);
            throw new SoarBridgeException("Error while Creaing SOAR Bridge");
        }
    }

    /**
     * Receive Commands from SOAR Kernel
     * @return A List of SOAR Commands
     * @throws SoarBridgeException
     */
    public ArrayList<SoarCommand> getReceivedCommands() throws SoarBridgeException
    {
        ArrayList<SoarCommand> commandList = new ArrayList<SoarCommand>();

        try
        {
            if (agent != null)
            {
                int numberCommands = agent.GetNumberCommands() ;

                for (int i = 0 ; i < numberCommands ; i++)
                {
                    Identifier pCommand = agent.GetCommand(i) ;
                    String name  = pCommand.GetCommandName();
                    SoarCommand.CommandType commandType = Enum.valueOf(SoarCommand.CommandType.class, name);
                    SoarCommand command = null;

                    switch(commandType)
                    {
                        case MOVE:
                            Float rightVelocity = null;
                            Float leftVelocity = null;
                            Float linearVelocity = null;
                            Float xPosition = null;
                            Float yPosition = null;

                            rightVelocity = tryParseFloat(pCommand.GetParameterValue("VelR"));
                            leftVelocity = tryParseFloat(pCommand.GetParameterValue("VelL"));
                            linearVelocity = tryParseFloat(pCommand.GetParameterValue("Vel"));
                            xPosition = tryParseFloat(pCommand.GetParameterValue("X"));
                            yPosition = tryParseFloat(pCommand.GetParameterValue("Y"));

                            command = new SoarCommand(SoarCommand.CommandType.MOVE);
                            SoarCommandMove commandMove = (SoarCommandMove)command.getCommandArgument();

                            if (commandMove != null)
                            {
                                if (rightVelocity != null)
                                {
                                    commandMove.setRightVelocity(rightVelocity);
                                }

                                if (leftVelocity != null)
                                {
                                    commandMove.setLeftVelocity(leftVelocity);
                                }

                                if (linearVelocity != null)
                                {
                                    commandMove.setLinearVelocity(linearVelocity);
                                }

                                if (xPosition != null)
                                {
                                    commandMove.setX(xPosition);
                                }

                                if (yPosition != null)
                                {
                                    commandMove.setY(yPosition);
                                }

                                commandList.add(command);
                            }
                            else
                            {
                                throw new NullPointerException("Command Move is null");
                            }

                            break;

                        case GET:
                            String thingNameToGet = null;

                            command = new SoarCommand(SoarCommand.CommandType.GET);
                            SoarCommandGet commandGet = (SoarCommandGet)command.getCommandArgument();

                            if (commandGet != null)
                            {
                                thingNameToGet = pCommand.GetParameterValue("Name");

                                if (thingNameToGet != null)
                                {
                                    commandGet.setThingName(thingNameToGet);
                                }

                                commandList.add(command);
                            }

                            break;

                        case EAT:
                            String thingNameToEat = null;

                            command = new SoarCommand(SoarCommand.CommandType.EAT);
                            SoarCommandEat commandEat = (SoarCommandEat)command.getCommandArgument();

                            if (commandEat != null)
                            {
                                thingNameToEat = pCommand.GetParameterValue("Name");

                                if (thingNameToEat != null)
                                {
                                    commandEat.setThingName(thingNameToEat);
                                }

                                commandList.add(command);
                            }

                            break;
                            
                        case DELIVERY:                            
                            command = new SoarCommand(SoarCommand.CommandType.DELIVERY);
                            commandList.add(command);
                            break;   
                            
                        default:
                            break;
                    }

                    // Update environment here to reflect agent's command
                    // Then mark the command as completed
                    pCommand.AddStatusComplete();
                }

                clearEntitiesToNextSimulation();
                agent.Commit();
                agent.ClearOutputLinkChanges();

                checkForKernelOrAgentError();
            }
        }
        catch (Exception e)
        {
            logger.error("Error while creating SOAR Kernel",e);
            throw new SoarBridgeException("Error while Creating SOAR Bridge");
        }

        return ((commandList.size() > 0) ? commandList : null);
    }

    /**
     * Check for error in SOAR Kernel or Agent
     * @throws SoarBridgeException
     */
    private void checkForKernelOrAgentError() throws SoarBridgeException
    {
        if (kernel != null && kernel.HadError())
        {
            throw new SoarBridgeException(kernel.GetLastErrorDescription());
        }

        if (agent != null && agent.HadError())
        {
            throw new SoarBridgeException(agent.GetLastErrorDescription());
        }
    }

    /**
     * Try Parse a Float Element
     * @param value Float Value
     * @return The Float Value or null otherwise
     */
    private Float tryParseFloat (String value)
    {
        Float returnValue = null;

        try
        {
            returnValue = Float.parseFloat(value);
        }
        catch (Exception ex)
        {
            returnValue = null;
        }

        return returnValue;
    }
    
    private String ProcessLeaflet(Vector<Leaflet> list)
    {        
        List<String> lista = null;
        lista = new ArrayList<String>();
        for (int i=0; i < list.size(); i++)
        {   
            Leaflet actual = list.get(i);
            HashMap<String, Integer[]> este = actual.getItems();
            for (Iterator<String> iter = este.keySet().iterator(); iter.hasNext();) 
            {
                String str = (String)iter.next(); //jewel color              
                int query = este.get(str)[0];
                int collected = este.get(str)[1];
                if(query > collected)
                {
                    lista.add(str);
                }
            }
        }
        
        if(lista.isEmpty())
        {  return ""; }
        else
        {
            int position = (int) Math.floor(Math.random()*lista.size());        
            String out = lista.get(position);
            return out;            
        }
    }
}
